@@ -21,27 +21,34 @@
  */
 
 #include "config.h"
-#include <stdlib.h>
-#include <string.h>
-#include <strings.h>
-#include <popt.h>
-#include <gdk/gdk.h>
-#include <gdk-pixbuf/gdk-pixbuf-io.h>
-#include <libbonobo.h>
-#include <login-helper/login-helper.h>
-#include <gdk/gdkx.h>
-#include <gtk/gtk.h>
-#include <X11/Xatom.h>
-#ifdef HAVE_XFIXES
-#include <X11/extensions/Xfixes.h>
-#endif
 #include "magnifier.h"
 #include "magnifier-private.h"
 #include "zoom-region.h"
 #include "zoom-region-private.h"
-#include "damage-client.h"
+#include "gmag-events.h"
 #include "GNOME_Magnifier.h"
 
+#include <stdlib.h>
+#include <string.h>
+#include <strings.h>
+
+#include <X11/Xatom.h>
+#ifdef HAVE_XFIXES
+#include <X11/extensions/Xfixes.h>
+#ifdef HAVE_COMPOSITE
+#include <X11/extensions/shape.h>
+#include <X11/extensions/Xcomposite.h>
+#endif /* HAVE_COMPOSITE */
+#endif /* HAVE_XFIXES */
+
+#include <libbonobo.h>
+#include <login-helper/login-helper.h>
+
+#include <gdk-pixbuf/gdk-pixbuf-io.h>
+#include <gdk/gdkx.h>
+#include <gdk/gdk.h>
+#include <gtk/gtk.h>
+
 /* if you #define this, don't forget to set MAG_CLIENT_DEBUG env variable */
 #define DEBUG_CLIENT_CALLS
 
@@ -133,12 +140,9 @@ gchar* mag_prop_names[MAGNIFIER_CROSSWIRE_COLOR_PROP + 1] = {
 #endif
 
 static int _x_error = 0;
-static int fixes_event_base = 0, fixes_error_base;
-static Display *cursor_client_connection;
-static guint    cursor_client_gsource = 0;
 static Magnifier *_this_magnifier = NULL;
+extern gint       fixes_event_base;
 
-static void magnifier_transform_cursor (Magnifier *magnifier);
 static void magnifier_init_cursor_set (Magnifier *magnifier, gchar *cursor_set);
 static void magnifier_init_window (Magnifier *magnifier, GdkScreen *screen);
 static gboolean magnifier_check_set_struts (Magnifier *magnifier);
@@ -235,9 +239,11 @@ can_open_display (gchar *display_name)
     Display *d;
     if ((d = XOpenDisplay (display_name)))
     {
+	    printf ("yes\n");
 	XCloseDisplay (d);
 	return TRUE;
     }
+    printf ("no\n");
     return FALSE;
 }
 
@@ -339,53 +345,6 @@ magnifier_set_cursor_from_pixbuf (Magnifier *magnifier, GdkPixbuf *cursor_pixbuf
 }
 
 
-void
-magnifier_free_cursor_pixels (guchar *pixels, gpointer data)
-{
-    /* XFree (data); FIXME why doesn't this work properly? */
-}
-
-#ifdef HAVE_XFIXES
-static void
-magnifier_cursor_convert_to_rgba (Magnifier *magnifier, XFixesCursorImage *cursor_image)
-{
-	int i, count = cursor_image->width * cursor_image->height;
-	for (i = 0; i < count; ++i) 
-	{
-		guint32 pixval = GUINT_TO_LE (cursor_image->pixels[i]);
-		cursor_image->pixels[i] = pixval;
-	}
-}
-#endif
-
-GdkPixbuf *
-magnifier_get_source_pixbuf (Magnifier *magnifier)
-{
-#ifdef HAVE_XFIXES
-	XFixesCursorImage *cursor_image = XFixesGetCursorImage (cursor_client_connection);
-        GdkPixbuf *cursor_pixbuf = NULL;
-	gchar s[6];
-	if (cursor_image)
-	{
-	        magnifier_cursor_convert_to_rgba (magnifier, cursor_image);
-		cursor_pixbuf = gdk_pixbuf_new_from_data ((guchar *) cursor_image->pixels, 
-							  GDK_COLORSPACE_RGB, 
-							  TRUE, 8,  
-							  cursor_image->width, cursor_image->height,
-							  cursor_image->width * 4, 
-							  magnifier_free_cursor_pixels, 
-							  cursor_image );
-		gdk_pixbuf_set_option (cursor_pixbuf, "x_hot", 
-				       g_ascii_dtostr (s, 6, (gdouble) cursor_image->xhot));
-		gdk_pixbuf_set_option (cursor_pixbuf, "y_hot", 
-				       g_ascii_dtostr (s, 6, (gdouble) cursor_image->yhot));
-	}
-	return cursor_pixbuf;
-#else
-	return NULL;
-#endif
-}
-
 GdkPixbuf *
 magnifier_get_pixbuf_for_name (Magnifier *magnifier, const gchar *cursor_name)
 {
@@ -398,7 +357,8 @@ magnifier_get_pixbuf_for_name (Magnifier *magnifier, const gchar *cursor_name)
 }
 
 void
-magnifier_set_cursor_pixmap_by_name (Magnifier *magnifier, const gchar *cursor_name, 
+magnifier_set_cursor_pixmap_by_name (Magnifier *magnifier,
+				     const gchar *cursor_name, 
 				     gboolean source_fallback)
 {
 	GdkPixbuf *pixbuf;
@@ -409,7 +369,7 @@ magnifier_set_cursor_pixmap_by_name (Magnifier *magnifier, const gchar *cursor_n
 #endif
 		if (source_fallback == TRUE)
 		{
-			pixbuf = magnifier_get_source_pixbuf (magnifier);
+			pixbuf = gmag_events_get_source_pixbuf (magnifier);
 		}
 		else
 		{
@@ -420,96 +380,6 @@ magnifier_set_cursor_pixmap_by_name (Magnifier *magnifier, const gchar *cursor_n
 	if (pixbuf) g_object_unref (pixbuf);
 }
 
-gboolean
-magnifier_cursor_notify (GIOChannel *source, GIOCondition condition, gpointer data)
-{
-#ifdef HAVE_XFIXES
-  XEvent ev;
-  Magnifier *magnifier = (Magnifier *) data;
-  XFixesCursorNotifyEvent  *cev = NULL;
-
-  /* TODO: don't ask for name unless xfixes version >= 2 available */
-  do
-  {
-      XNextEvent(cursor_client_connection, &ev);
-      if (ev.type == fixes_event_base + XFixesCursorNotify) 
-      {
-	  cev = (XFixesCursorNotifyEvent *) &ev;
-      }
-  } while (XPending (cursor_client_connection));
-
-  if (magnifier->priv->use_source_cursor) 
-  {
-      GdkPixbuf *cursor_pixbuf = magnifier_get_source_pixbuf (magnifier);
-      magnifier_set_cursor_from_pixbuf (magnifier, cursor_pixbuf);
-      if (cursor_pixbuf) g_object_unref (cursor_pixbuf);
-  }
-  else
-  {
-      magnifier_set_cursor_pixmap_by_name (magnifier, 
-					   cev ? gdk_x11_get_xatom_name (cev->cursor_name) : "default",
-					   TRUE);
-  }
-
-  magnifier_transform_cursor (magnifier);
-#ifdef DEBUG_CURSOR
-  if (cev)
-  g_message ("cursor changed: subtype=%d, cursor_serial=%lu, name=[%x] %s\n",
-	     (int) cev->subtype, cev->cursor_serial, (int) cev->cursor_name,
-	     gdk_x11_get_xatom_name (cev->cursor_name));
-#endif
-
-  return TRUE;
-#else
-  return FALSE;
-#endif
-}
-
-gboolean
-magnifier_cursor_notification_init (Magnifier *magnifier)
-{
-#ifdef HAVE_XFIXES
-    GIOChannel *ioc;
-    int fd;
-    Window rootwin;
-
-    if (!magnifier->source_display) return FALSE;
-
-    if (cursor_client_connection)
-    {
-        /* remove the old watch */
-	if (cursor_client_gsource) 
-	  g_source_remove (cursor_client_gsource);
-	XCloseDisplay (cursor_client_connection);
-    }
-
-    cursor_client_connection = XOpenDisplay (magnifier->source_display_name);
-    rootwin = GDK_WINDOW_XWINDOW (magnifier->priv->root);
-
-    if (!XFixesQueryExtension (cursor_client_connection, &fixes_event_base, &fixes_error_base))
-    {
-	g_warning ("XFixes extension not currently active.\n");
-	return FALSE;
-    }
-    else
-    {
-	XFixesSelectCursorInput (cursor_client_connection, rootwin, XFixesDisplayCursorNotifyMask);
-	fd = ConnectionNumber (cursor_client_connection);
-	ioc = g_io_channel_unix_new (fd);
-	cursor_client_gsource = 
-	    g_io_add_watch (ioc, G_IO_IN | G_IO_HUP | G_IO_PRI | G_IO_ERR, magnifier_cursor_notify, 
-			    magnifier);
-	g_io_channel_unref (ioc); 
-	g_message ("added event source to xfixes cursor-notify connection");
- 	XFlush (cursor_client_connection); 
-   }
-    return TRUE;
-#else
-    g_warning ("this copy of gnome-mag was built without xfixes extension support.\n");
-    return FALSE;
-#endif
-}
-
 void
 magnifier_notify_damage (Magnifier *magnifier, XRectangle *rect)
 {
@@ -529,10 +399,7 @@ magnifier_notify_damage (Magnifier *magnifier, XRectangle *rect)
 static void
 magnifier_set_extension_listeners (Magnifier *magnifier, GdkWindow *root)
 {
-	if (!magnifier_damage_client_init (magnifier))
-	    g_warning ("Damage client hooks were not initialized.");
-	if (!magnifier_cursor_notification_init (magnifier))
-	    g_warning ("Cursor change notification hooks were not initialized.");
+	gmag_events_client_init (magnifier);
 	magnifier->source_initialized = TRUE;
 }
 
@@ -664,6 +531,11 @@ magnifier_adjust_source_size (Magnifier *magnifier)
 	hfract_right = (double) (rect_bounds.x2 - magnifier->target_bounds.x2) / (double) rect_bounds.x2;
 	vfract_bottom = (double) (rect_bounds.y2 - magnifier->target_bounds.y2) / (double) rect_bounds.y2;
 	/* we make our 'source' rectangle the largest available subsection which we aren't occupying */
+#ifdef HAVE_COMPOSITE
+	if (!g_getenv ("MAGNIFIER_IGNORE_COMPOSITE")) {
+		magnifier->source_bounds = rect_bounds;
+	} else
+#endif
 	if (MAX (hfract_left, hfract_right) > MAX (vfract_top, vfract_bottom))  /* vertical split, approximately */
 	{
 		if (hfract_right > hfract_left) {
@@ -890,7 +762,7 @@ magnifier_recolor_pixbuf (Magnifier *magnifier, GdkPixbuf *pixbuf)
 	}
 }
 
-static void
+void
 magnifier_transform_cursor (Magnifier *magnifier)
 {
 	if (magnifier->priv->cursor) /* don't do this if cursor isn't intialized yet */
@@ -1051,6 +923,51 @@ magnifier_init_cursor_set (Magnifier *magnifier, gchar *cursor_set)
 	magnifier_transform_cursor (magnifier);
 }
 
+static gboolean
+magnifier_reset_overlay_at_idle (gpointer data)
+{
+#ifdef HAVE_COMPOSITE
+	if (!g_getenv ("MAGNIFIER_IGNORE_COMPOSITE")) {
+		if (data) {
+			Magnifier *magnifier = MAGNIFIER (data);
+			if (magnifier->priv &&
+			    GTK_WIDGET_REALIZED (magnifier->priv->w)) 
+				if (magnifier->priv->w->window) {
+#ifdef HAVE_OVERLAY
+					gint      x, y, over_w, over_h;
+
+					magnifier->priv->overlay = gdk_window_foreign_new (XCompositeGetOverlayWindow (GDK_WINDOW_XDISPLAY (magnifier->priv->w->window), GDK_WINDOW_XID (magnifier->priv->root)));
+					gdk_window_get_position (
+						magnifier->priv->w->window,
+						&x, &y);
+					gdk_drawable_get_size (
+						magnifier->priv->w->window,
+						&over_w, &over_h);
+					gdk_window_move_resize (
+						magnifier->priv->overlay,
+						x, y, over_w, over_h);
+					gdk_window_reparent (
+						magnifier->priv->w->window,
+						magnifier->priv->overlay,
+						0, 0);
+#ifdef HAVE_SHAPEINPUT			
+					XShapeCombineRectangles (GDK_WINDOW_XDISPLAY (magnifier->priv->w->window), GDK_WINDOW_XID (magnifier->priv->overlay), ShapeInput, 0, 0, NULL, 0, ShapeSet, YXBanded);
+#endif /* HAVE_SHAPEINPUT */
+#else
+#ifdef HAVE_SHAPEINPUT
+					XShapeCombineRectangles (GDK_WINDOW_XDISPLAY (magnifier->priv->w->window), GDK_WINDOW_XID (magnifier->priv->w->window), ShapeInput, 0, 0, NULL, 0, ShapeSet, YXBanded);
+#endif /* HAVE_SHAPEINPUT */
+#endif /* HAVE_OVERLAY */
+					return FALSE;
+				}
+		}
+	} else
+		return FALSE;
+	return TRUE;
+#endif
+	return FALSE;
+}
+
 static gboolean 
 magnifier_reset_struts_at_idle (gpointer data)
 {
@@ -1069,7 +986,11 @@ magnifier_reset_struts_at_idle (gpointer data)
 static gboolean
 magnifier_check_set_struts (Magnifier *magnifier)
 {
-	/* TODO: don't do this if we're using Composite */
+#ifdef HAVE_COMPOSITE
+	if (!g_getenv ("MAGNIFIER_IGNORE_COMPOSITE"))
+		return TRUE;
+#endif
+
 	if (magnifier &&
 	    magnifier->priv && magnifier->priv->w && GTK_WIDGET_REALIZED (magnifier->priv->w) &&
 	    magnifier->priv->w->window) 
@@ -1236,8 +1157,10 @@ magnifier_set_property (BonoboPropertyBag *bag,
 	switch (arg_id) {
 	case MAGNIFIER_SOURCE_DISPLAY_PROP:
 		full_display_string = BONOBO_ARG_GET_STRING (arg);
+		printf ("%s\n", full_display_string);
 		if (can_open_display (full_display_string))
 		{
+			printf ("here\n");
 		    GSList *zoom_region_params = NULL;
 		    magnifier->source_screen_num =
 			magnifier_parse_display_name (magnifier,
@@ -1865,6 +1788,8 @@ magnifier_init (Magnifier *magnifier)
 	magnifier->priv->w = NULL;
 	magnifier->priv->use_source_cursor = TRUE;
 	magnifier->priv->cursorlist = NULL;
+	magnifier->priv->source_drawable = NULL;
+	magnifier->priv->overlay = NULL;
 	magnifier_init_window (magnifier, 
 			       gdk_display_get_screen (magnifier->target_display, 
 						       magnifier->target_screen_num));
@@ -1900,7 +1825,8 @@ magnifier_get_cursor (Magnifier *magnifier)
 	    }
 #ifdef HAVE_XFIXES
 	    else {
-		GdkPixbuf *cursor_pixbuf = magnifier_get_source_pixbuf (magnifier);
+		GdkPixbuf *cursor_pixbuf = gmag_events_get_source_pixbuf (
+			magnifier);
 		magnifier_set_cursor_from_pixbuf (magnifier, cursor_pixbuf);
 		if (cursor_pixbuf) g_object_unref (cursor_pixbuf);
 		magnifier_transform_cursor (magnifier);
@@ -1935,6 +1861,7 @@ magnifier_new (gboolean override_redirect)
 	    g_error ("Error registering magnifier server.\n");
 
 	g_idle_add (magnifier_reset_struts_at_idle, mag);
+	g_idle_add (magnifier_reset_overlay_at_idle, mag);
 
 	return mag;
 }
